//Manually written Wrapper for C web services
//01 September 2003 by Susantha Kumara (susantha@opensource.lk, skumara@virtusa.com)

#include "CServiceWrapper.hpp"

extern "C" int AddInt(int a, int b);
extern "C" double Distance(Point *p1, Point *p2);
extern "C" Point* AddPoint(Point *p1, Point *p2);
extern "C" double Perimeter(const Triangle* pTri);

//Parameters and wrapper methos to manipulate Point
static const AxisChar* Axis_URI_Point = "http://www.opensource.lk/Point";
static const AxisChar* Axis_TypeName_Point = "Point";

int Axis_DeSerialize_Point(Point* p, IWrapperSoapDeSerializer *pDZ)
{
	p->x = pDZ->GetParam()->GetInt();
	p->y = pDZ->GetParam()->GetInt();
	return SUCCESS;
}

void Axis_Delete_Point(Point* p, bool bArray = false, int nSize=0)
{
	if (bArray)
		delete [] p;
	else
		delete p;
}

int Axis_Serialize_Point(Point* p, IWrapperSoapSerializer& pSZ, bool bArray = false)
{
	if (bArray)
	{
		pSZ << "<" << Axis_TypeName_Point << ">";
	}
	else
	{
		const AxisChar* sPrefix = pSZ.getNewNamespacePrefix();
		pSZ << "<" << Axis_TypeName_Point << " xsi:type=\"" << sPrefix <<":" 
			<< Axis_TypeName_Point << " xmlns:" << sPrefix << "=\"" 
			<< Axis_URI_Point << "\">";
	}
	pSZ << pSZ.SerializeBasicType("x", p->x, XSD_INT);
	pSZ << pSZ.SerializeBasicType("y", p->y, XSD_INT);
	pSZ << "</" << Axis_TypeName_Point << ">";
	return SUCCESS;
}

int Axis_GetSize_Point(Point* p)
{
	return sizeof(Point);
}

//Parameters and wrapper methods to manipulate Triangle
static const AxisChar* Axis_URI_Triangle = "http://www.opensource.lk/Triangle";
static const AxisChar* Axis_TypeName_Triangle = "Triangle";

int Axis_DeSerialize_Triangle(Triangle* p, IWrapperSoapDeSerializer *pDZ)
{
	IParam* pParam;
	pParam = pDZ->GetParam(); //get head param and do any validation if necessary.
	p->p1 = new Point;
	Axis_DeSerialize_Point(p->p1, pDZ);
	pParam = pDZ->GetParam(); //get head param and do any validation if necessary.
	p->p2 = new Point;
	Axis_DeSerialize_Point(p->p2, pDZ);
	pParam = pDZ->GetParam(); //get head param and do any validation if necessary.
	p->p3 = new Point;
	Axis_DeSerialize_Point(p->p3, pDZ);
	return SUCCESS;
}

void Axis_Delete_Triangle(Triangle* p, bool bArray = false, int nSize=0)
{
	if (bArray)
	{
		//delete any pointer members of this struct here
		Triangle* pTemp = p;
		for (int x=0; x<nSize; x++)
		{
			Axis_Delete_Point(pTemp->p1);
			Axis_Delete_Point(pTemp->p2);
			Axis_Delete_Point(pTemp->p3);
			pTemp++;
		}
		delete [] p;
	}
	else
	{
		//delete any pointer members of this struct here
		Axis_Delete_Point(p->p1);
		Axis_Delete_Point(p->p2);
		Axis_Delete_Point(p->p3);
		delete p;
	}
}

int Axis_Serialize_Triangle(Triangle* p, IWrapperSoapSerializer& pSZ, bool bArray = false)
{
	if (bArray)
	{
		pSZ << "<" << Axis_TypeName_Triangle << ">";
	}
	else
	{
		AxisString sPrefix = pSZ.getNewNamespacePrefix();
		pSZ << "<" << Axis_TypeName_Triangle << " xsi:type=\"" << sPrefix.c_str() <<":" 
			<< Axis_TypeName_Triangle << " xmlns:" << sPrefix.c_str() << "=\"" 
			<< Axis_URI_Triangle << "\">";
	}
	Axis_Serialize_Point(p->p1, pSZ);
	Axis_Serialize_Point(p->p1, pSZ);
	Axis_Serialize_Point(p->p1, pSZ);
	pSZ << "</" << Axis_TypeName_Triangle << ">";
	return SUCCESS;
}

int Axis_GetSize_Triangle(Triangle* p)
{
	return sizeof(Triangle);
}

CServiceWrapper::CServiceWrapper()
{
	//does nothing
}

CServiceWrapper::~CServiceWrapper()
{
	//does nothing
}

// Implementation of WrapperClassHandler interfaces.
int CServiceWrapper::Invoke(IMessageData* mc)
{
	IWrapperSoapDeSerializer* pIWSDZ = NULL;
	mc->getSoapDeSerializer(&pIWSDZ);
	if (!pIWSDZ) return FAIL;
	const AxisChar* method = pIWSDZ->GetMethodName();
	if (0 == strcmp(method, "Distance"))
		return DistanceWrapped(mc);
	else if (0 == strcmp(method, "AddPoint"))
		return AddPointWrapped(mc);
	else if (0 == strcmp(method, "Add"))
		return AddWrapped(mc);
	else if (0 == strcmp(method, "Perimeter"))
		return PerimeterWrapped(mc);
	else return FAIL;
}

void CServiceWrapper::OnFault(IMessageData* pMsg)
{
}

int CServiceWrapper::Init()
{
	return 0;
}

int CServiceWrapper::Fini()
{
	return 0;
}

int CServiceWrapper::DistanceWrapped(IMessageData* mc)
{
	IWrapperSoapSerializer* pIWSSZ = NULL;
	mc->getSoapSerializer(&pIWSSZ);
	if (!pIWSSZ) return FAIL;
	IWrapperSoapDeSerializer* pIWSDZ = NULL;
	mc->getSoapDeSerializer(&pIWSDZ);
	if (!pIWSDZ) return FAIL;
	pIWSSZ->createSoapMethod("DistanceResponse", pIWSSZ->getNewNamespacePrefix(), "http://www.opensource.lk/CPPService");

	IParam *param0 = pIWSDZ->GetParam();
	Point* v0 = new Point();

	param0->SetUserType(v0, (AXIS_DESERIALIZE_FUNCT)Axis_DeSerialize_Point, (AXIS_OBJECT_DELETE_FUNCT)Axis_Delete_Point);
	pIWSDZ->Deserialize(param0,0);

	IParam *param1 = pIWSDZ->GetParam();
	Point* v1 = new Point();

	param1->SetUserType(v1, (AXIS_DESERIALIZE_FUNCT)Axis_DeSerialize_Point, (AXIS_OBJECT_DELETE_FUNCT)Axis_Delete_Point);
	pIWSDZ->Deserialize(param1,0);

	//Call actual web service method with appropriate parameters
	double ret = Distance(v0, v1);

	return pIWSSZ->AddOutputParam("DistanceReturn", ret, XSD_DOUBLE);
}

int CServiceWrapper::AddPointWrapped(IMessageData* mc)
{
	IWrapperSoapSerializer* pIWSSZ = NULL;
	mc->getSoapSerializer(&pIWSSZ);
	if (!pIWSSZ) return FAIL;
	IWrapperSoapDeSerializer* pIWSDZ = NULL;
	mc->getSoapDeSerializer(&pIWSDZ);
	if (!pIWSDZ) return FAIL;
	pIWSSZ->createSoapMethod("AddPointResponse", pIWSSZ->getNewNamespacePrefix(), "http://www.opensource.lk/CPPService");

	IParam *param0 = pIWSDZ->GetParam();
	Point* v0 = new Point();

	param0->SetUserType(v0, (AXIS_DESERIALIZE_FUNCT)Axis_DeSerialize_Point, (AXIS_OBJECT_DELETE_FUNCT)Axis_Delete_Point);
	pIWSDZ->Deserialize(param0,0);

	IParam *param1 = pIWSDZ->GetParam();
	Point* v1 = new Point();

	param1->SetUserType(v1, (AXIS_DESERIALIZE_FUNCT)Axis_DeSerialize_Point, (AXIS_OBJECT_DELETE_FUNCT)Axis_Delete_Point);
	pIWSDZ->Deserialize(param1,0);

	//Call actual web service method with appropriate parameters
	Point* ret = AddPoint(v0, v1);

	return pIWSSZ->AddOutputParam("AddPointReturn", ret, (void*)Axis_Serialize_Point, (void*)Axis_Delete_Point);
}

int CServiceWrapper::AddWrapped(IMessageData* mc)
{
	IWrapperSoapSerializer* pIWSSZ = NULL;
	mc->getSoapSerializer(&pIWSSZ);
	if (!pIWSSZ) return FAIL;
	IWrapperSoapDeSerializer* pIWSDZ = NULL;
	mc->getSoapDeSerializer(&pIWSDZ);
	if (!pIWSDZ) return FAIL;
	pIWSSZ->createSoapMethod("AddResponse", pIWSSZ->getNewNamespacePrefix(), "http://www.opensource.lk/CPPService");

	IParam *param0 = pIWSDZ->GetParam();
	int v0 = param0->GetInt();

	IParam *param1 = pIWSDZ->GetParam();
	int v1 = param1->GetInt();

	//Call actual web service method with appropriate parameters
	int ret = AddInt(v0, v1);

	return pIWSSZ->AddOutputParam("AddReturn", ret, XSD_INT);
}

int CServiceWrapper::PerimeterWrapped(IMessageData* mc)
{
	IWrapperSoapSerializer* pIWSSZ = NULL;
	mc->getSoapSerializer(&pIWSSZ);
	if (!pIWSSZ) return FAIL;
	IWrapperSoapDeSerializer* pIWSDZ = NULL;
	mc->getSoapDeSerializer(&pIWSDZ);
	if (!pIWSDZ) return FAIL;
	pIWSSZ->createSoapMethod("PerimeterResponse", pIWSSZ->getNewNamespacePrefix(), "http://www.opensource.lk/CPPService");

	IParam *param0 = pIWSDZ->GetParam();
	Triangle* v0 = new Triangle();

	param0->SetUserType(v0, (AXIS_DESERIALIZE_FUNCT)Axis_DeSerialize_Triangle, (AXIS_OBJECT_DELETE_FUNCT)Axis_Delete_Triangle);
	pIWSDZ->Deserialize(param0,0);

	//Call actual web service method with appropriate parameters
	double ret = Perimeter(v0);

	return pIWSSZ->AddOutputParam("PerimeterReturn", ret, XSD_DOUBLE);
}
